# MIT License

# Copyright (c) 2016, Tin Tvrtković

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
# (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, 
# publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
# FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from collections import defaultdict
from typing import Any, Callable, Dict, Literal, Optional, Type, Union

from attr import NOTHING

from cattrs import BaseConverter
from cattrs._compat import get_newtype_base, is_literal, is_subclass, is_union_type


def default_tag_generator(typ: Type) -> str:
    """Return the class name."""
    return typ.__name__


def configure_tagged_union(
    union: Any,
    converter: BaseConverter,
    tag_generator: Callable[[Type], str] = default_tag_generator,
    tag_name: str = "_type",
    default: Optional[Type | Literal[NOTHING]] = NOTHING,
) -> None:
    """
    Configure the converter so that `union` (which should be a union) is
    un/structured with the help of an additional piece of data in the
    unstructured payload, the tag.

    :param converter: The converter to apply the strategy to.
    :param tag_generator: A `tag_generator` function is used to map each
        member of the union to a tag, which is then included in the
        unstructured payload. The default tag generator returns the name of
        the class.
    :param tag_name: The key under which the tag will be set in the
        unstructured payload. By default, `'_type'`.
    :param default: An optional class to be used if the tag information
        is not present when structuring.

    The tagged union strategy currently only works with the dict
    un/structuring base strategy.

    .. versionadded:: 23.1.0
    """
    args = union.__args__
    tag_to_hook = {}
    exact_cl_unstruct_hooks = {}
    for cl in args:
        tag = tag_generator(cl)
        struct_handler = converter._structure_func.dispatch(cl)
        unstruct_handler = converter._unstructure_func.dispatch(cl)

        def structure_union_member(val: dict, _cl=cl, _h=struct_handler) -> cl:
            return _h(val, _cl)

        def unstructure_union_member(val: union, _h=unstruct_handler) -> dict:
            return _h(val)

        tag_to_hook[tag] = structure_union_member
        exact_cl_unstruct_hooks[cl] = unstructure_union_member

    cl_to_tag = {cl: tag_generator(cl) for cl in args}

    if default is not NOTHING:
        default_handler = converter._structure_func.dispatch(default)

        def structure_default(val: dict, _cl=default, _h=default_handler):
            return _h(val, _cl)

        tag_to_hook = defaultdict(lambda: structure_default, tag_to_hook)
        cl_to_tag = defaultdict(lambda: default, cl_to_tag)

    def unstructure_tagged_union(
        val: union,
        _exact_cl_unstruct_hooks=exact_cl_unstruct_hooks,
        _cl_to_tag=cl_to_tag,
        _tag_name=tag_name,
    ) -> Dict:
        res = _exact_cl_unstruct_hooks[val.__class__](val)
        res[_tag_name] = _cl_to_tag[val.__class__]
        return res

    if default is NOTHING:

        def structure_tagged_union(
            val: dict, _, _tag_to_cl=tag_to_hook, _tag_name=tag_name
        ) -> union:
            return _tag_to_cl[val[_tag_name]](val)

    else:

        def structure_tagged_union(
            val: dict,
            _,
            _tag_to_hook=tag_to_hook,
            _tag_name=tag_name,
            _dh=default_handler,
            _default=default,
        ) -> union:
            if _tag_name in val:
                return _tag_to_hook[val[_tag_name]](val)
            else:
                return _dh(val, _default)

    converter.register_unstructure_hook(union, unstructure_tagged_union)
    converter.register_structure_hook(union, structure_tagged_union)



def configure_union_passthrough(
    union: Any, converter: BaseConverter, accept_ints_as_floats: bool = True
) -> None:
    """
    Configure the converter to support validating and passing through unions of the
    provided types and their subsets.

    For example, all mature JSON libraries natively support producing unions of ints,
    floats, Nones, and strings. Using this strategy, a converter can be configured
    to efficiently validate and pass through unions containing these types.

    The most important point is that another library (in this example the JSON
    library) handles producing the union, and the converter is configured to just
    validate it.

    Literals of provided types are also supported, and are checked by value.

    NewTypes of provided types are also supported.

    The strategy is designed to be O(1) in execution time, and independent of the
    ordering of types in the union.

    If the union contains a class and one or more of its subclasses, the subclasses
    will also be included when validating the superclass.

    :param accept_ints_as_floats: When set (the default), if the provided union
        contains both ints and floats, actual unions containing only floats will also accept
        ints. See https://typing.python.org/en/latest/spec/special-types.html#special-cases-for-float-and-complex
        for more information.
    """
    args = set(union.__args__)

    def make_structure_native_union(exact_type: Any) -> Callable:
        # `exact_type` is likely to be a subset of the entire configured union (`args`).
        literal_values = {
            v for t in exact_type.__args__ if is_literal(t) for v in t.__args__
        }

        # We have no idea what the actual type of `val` will be, so we can't
        # use it blindly with an `in` check since it might not be hashable.
        # So we do an additional check when handling literals.
        # Note: do no use `literal_values` here, since {0, False} gets reduced to {0}
        literal_classes = {
            v.__class__
            for t in exact_type.__args__
            if is_literal(t)
            for v in t.__args__
        }

        non_literal_classes = {
            get_newtype_base(t) or t
            for t in exact_type.__args__
            if not is_literal(t) and ((get_newtype_base(t) or t) in args)
        }

        # We augment the set of allowed classes with any configured subclasses of
        # the exact subclasses.
        non_literal_classes |= {
            a for a in args if any(is_subclass(a, c) for c in non_literal_classes)
        }

        # We check for spillover - union types not handled by the strategy.
        # If spillover exists and we fail to validate our types, we call
        # further into the converter with the rest.
        spillover = {
            a
            for a in exact_type.__args__
            if (get_newtype_base(a) or a) not in non_literal_classes
            and not is_literal(a)
        }

        # By default, when floats are part of the union, accept ints too.
        if (
            accept_ints_as_floats
            and int in args
            and float in args
            and float in non_literal_classes
            and int not in non_literal_classes
        ):
            non_literal_classes.add(int)

        if spillover:
            spillover_type = (
                Union[tuple(spillover)] if len(spillover) > 1 else next(iter(spillover))
            )

            def structure_native_union(
                val: Any,
                _: Any,
                classes=non_literal_classes,
                vals=literal_values,
                converter=converter,
                spillover=spillover_type,
            ) -> exact_type:
                if val.__class__ in literal_classes and val in vals:
                    return val
                if val.__class__ in classes:
                    return val
                return converter.structure(val, spillover)

        else:

            def structure_native_union(
                val: Any, _: Any, classes=non_literal_classes, vals=literal_values
            ) -> exact_type:
                if val.__class__ in literal_classes and val in vals:
                    return val
                if val.__class__ in classes:
                    return val
                raise TypeError(f"{val} ({val.__class__}) not part of {_}")

        return structure_native_union

    def contains_native_union(exact_type: Any) -> bool:
        """Can we handle this type?"""
        if is_union_type(exact_type):
            type_args = set(exact_type.__args__)
            # We special case optionals, since they are very common
            # and are handled a little more efficiently by default.
            if len(type_args) == 2 and type(None) in type_args:
                return False

            literal_classes = {
                lit_arg.__class__
                for t in type_args
                if is_literal(t)
                for lit_arg in t.__args__
            }
            non_literal_types = {
                get_newtype_base(t) or t for t in type_args if not is_literal(t)
            }

            return (literal_classes | non_literal_types) & args
        return False

    converter.register_structure_hook_factory(
        contains_native_union, make_structure_native_union
    )


# Copyright 2025 Cyber Skyline

# Permission is hereby granted, free of charge, to any person obtaining a 
# copy of this software and associated documentation files (the “Software”), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
# IN THE SOFTWARE.
from cyber_skyline.chall_parser.compose.compose import ComposeFile, Network, ComposeResourceName, ServicesDict, NetworksDict
from cyber_skyline.chall_parser.compose.service import Service
from cyber_skyline.chall_parser.compose.challenge_info import ChallengeInfo, TextBody, Hint, Question, Variable

__all__ = [
    'ComposeFile',
    'Service',
    'ChallengeInfo',
    'TextBody',
    'Hint',
    'Question',
    'Variable',
    'Network',
    'ComposeResourceName',
    'ServicesDict',
    'NetworksDict'
]



# Copyright 2025 Cyber Skyline

# Permission is hereby granted, free of charge, to any person obtaining a 
# copy of this software and associated documentation files (the “Software”), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
# IN THE SOFTWARE.
from typing import cast
import pytest
from cyber_skyline.chall_parser.compose import (
    ComposeFile, Network, ComposeResourceName, ServicesDict, NetworksDict, Service, ChallengeInfo
    
)
from cyber_skyline.chall_parser.compose.validators import validate_compose_name_pattern

class TestNetwork:
    def test_network_requirements(self):
        """Test that networks must be internal."""
        network = Network(internal=True)
        assert network.internal is True

class TestComposeFile:
    def test_minimal_compose_file(self):
        """Test creating a minimal compose file."""
        challenge = ChallengeInfo(
            name="Test Challenge",
            description="A test challenge",
            questions=[]
        )
        compose = ComposeFile(challenge=challenge)
        assert compose.challenge.name == "Test Challenge"
        assert compose.services is None
        assert compose.networks is None

    def test_compose_file_with_all_sections(self):
        """Test ComposeFile with all possible sections."""
        challenge = ChallengeInfo(
            name="Full Test",
            description="Testing all sections",
            questions=[]
        )
        
        services = {
            ComposeResourceName("web"): Service(image="nginx", hostname="web"),
            ComposeResourceName("db"): Service(image="postgres", hostname="db")
        }
        
        networks = {
            ComposeResourceName("frontend"): Network(internal=True),
            ComposeResourceName("backend"): Network(internal=True)
        }
        
        compose = ComposeFile(
            challenge=challenge,
            services=cast(ServicesDict, services),
            networks=cast(NetworksDict, networks)
        )
        
        assert len(compose.services) == 2
        assert compose.networks is not None
        assert len(compose.networks) == 2
        assert compose.challenge.name == "Full Test"

class TestComposeNameValidation:
    def test_compose_name_pattern_validator_edge_cases(self):
        """Test edge cases for compose name pattern validation."""
        # Test with empty dict
        validate_compose_name_pattern(None, type('obj', (), {'name': 'test'})(), {})
        
        # Test with valid patterns
        valid_names = {
            "a": "value",
            "1": "value", 
            "a-b": "value",
            "a_b": "value",
            "a.b": "value",
            "a1b2c3": "value"
        }
        validate_compose_name_pattern(None, type('obj', (), {'name': 'test'})(), valid_names)

    def test_invalid_compose_names(self):
        """Test that invalid compose names are rejected."""
        invalid_dict = {
            "service with spaces": "value",
            "service@special": "value"
        }
        
        class MockInstance:
            pass
        
        class MockAttribute:
            name = "services"
        
        with pytest.raises(ValueError, match="must match pattern"):
            validate_compose_name_pattern(MockInstance(), MockAttribute(), invalid_dict)


# Copyright 2025 Cyber Skyline

# Permission is hereby granted, free of charge, to any person obtaining a 
# copy of this software and associated documentation files (the “Software”), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
# IN THE SOFTWARE.
import attr
import attr.validators as v

@attr.s
class AnswerTestCase:
    """Represents a test case for validating an answer.
    
    Each test case can be used to check if the answer is correct.
    """
    answer: str = attr.ib(validator=v.instance_of(str))
    """The expected answer text for this test case."""
    correct: bool = attr.ib(validator=v.instance_of(bool))
    """Indicates if this test case is a correct answer.
    If True, the answer is expected to match this test case.
    If False, the answer should not match this test case.
    """


@attr.s
class Answer:
    body: str = attr.ib(validator=v.instance_of(str))  # The regex pattern
    test_cases: list[AnswerTestCase] | None = attr.ib(default=None, validator=v.optional(v.deep_iterable(v.instance_of(AnswerTestCase), v.instance_of(list))))  # Optional test cases for validating the answer




# Copyright 2025 Cyber Skyline

# Permission is hereby granted, free of charge, to any person obtaining a 
# copy of this software and associated documentation files (the “Software”), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
# IN THE SOFTWARE.
import logging
import yaml
from faker import Faker
from faker.providers import address, automotive, bank, barcode, \
                            color, company, credit_card, currency, \
                            date_time, doi, emoji, file, geo, internet, \
                            isbn, job, lorem, misc, passport, person, \
                            phone_number, profile, python, sbn, ssn, user_agent

logger = logging.getLogger(__name__)

fake = Faker()

all_providers = [
    address, automotive, bank, barcode, 
    color, company, credit_card, currency, 
    date_time, doi, emoji, file, geo, internet, 
    isbn, job, lorem, misc, passport, person, 
    phone_number, profile, python, sbn, ssn, user_agent
]

# Register all Faker providers
for provider in all_providers:
    fake.add_provider(provider)

class Template:
    def __init__(self, eval_str: str, parent_variable: str):
        """Initialize a Template object with a template string and optional parent variable."""
        logger.debug(f"Initializing Template with template: {eval_str}, parent_variable: {parent_variable}")
        self.eval_str: str = eval_str
        self.parent_variable: str = parent_variable
    
    def eval(self, seed: str | None = None):
        """Evaluate the template using Faker."""
        if seed is not None:
            fake.seed_instance(seed)

        logger.debug(f"Evaluating template: {self.eval_str}")
        evaluated = None
        if not self.eval_str:
            logger.debug("Template is empty, returning None")
            return None
        try:
            # Use eval to execute the template code
            evaluated = eval(self.eval_str, {'fake': fake})
        except Exception as e:
            logger.error(f"Error evaluating template '{self.eval_str}': {e}")
            raise ValueError(f"Invalid template: {self.eval_str}") from e
        
        return evaluated


    @classmethod
    def from_yaml(cls, loader: yaml.Loader | yaml.FullLoader | yaml.UnsafeLoader | None, node: yaml.nodes.Node) -> 'Template':
        logger.debug(f"Creating Template from YAML node: {node}")
        if not isinstance(node, yaml.nodes.MappingNode):
            raise yaml.YAMLError("Template must be a mapping")
        
        values = {k.value: v for k, v in node.value}
        if 'eval' not in values:
            raise yaml.YAMLError("Template must have a 'eval' key")
        if 'variable' not in values:
            raise yaml.YAMLError("Template must have a 'variable' key")
        
        return cls(eval_str=values['eval'].value, parent_variable=values['variable'].value)
    
    @classmethod
    def to_yaml(cls, dumper: yaml.Dumper, data: 'Template') -> yaml.nodes.MappingNode:
        logger.debug(f"Converting Template to YAML: {data}")
        return dumper.represent_mapping("!template", [
            ('eval', data.eval_str),
            ('variable', data.parent_variable)
        ])


# Copyright 2025 Cyber Skyline

# Permission is hereby granted, free of charge, to any person obtaining a 
# copy of this software and associated documentation files (the “Software”), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
# IN THE SOFTWARE.
import pytest
import yaml
import io
from yaml import BaseLoader
from cyber_skyline.chall_parser.rewriter import Rewriter
from cyber_skyline.chall_parser.template import Template

class TestRewriter:
    def test_rewriter_initialization(self):
        """Test that Rewriter initializes correctly."""
        yaml_content = "test: value"
        loader = BaseLoader(yaml_content)
        rewriter = Rewriter(loader)
        assert rewriter.loader == loader
        assert rewriter.anchors == {}

    def test_rewrite_variable_no_anchor_in_default(self):
        """Test that rewriter fails when default has no anchor."""
        yaml_content = """
variables:
  bad_var:
    template: "fake.word()"
    default: "no_anchor_value"  # missing anchor
"""
        loader = BaseLoader(yaml_content)
        rewriter = Rewriter(loader)
        
        # This should raise an error because anchor is required for rewriting
        with pytest.raises(ValueError, match="Default value must have an anchor"):
            list(rewriter.rewrite())

    def test_rewrite_without_variables_section(self):
        """Test rewriting YAML without variables section."""
        yaml_content = """
services:
  web:
    image: nginx
"""
        loader = BaseLoader(yaml_content)
        rewriter = Rewriter(loader)
        events = list(rewriter.rewrite())
        # Should pass through all events unchanged
        assert len(events) > 0

    def test_rewrite_empty_variables_section(self):
        """Test rewriting YAML with empty variables section."""
        yaml_content = """
variables:
services:
  web:
    image: nginx
"""
        loader = BaseLoader(yaml_content)
        rewriter = Rewriter(loader)
        events = list(rewriter.rewrite())
        # Should handle empty variables section
        assert len(events) > 0

    def test_resolve_alias_not_found(self):
        """Test alias resolution when anchor is not found."""
        yaml_content = "test: value"
        loader = BaseLoader(yaml_content)
        rewriter = Rewriter(loader)
        
        # Create a mock alias event
        from yaml import AliasEvent
        alias = AliasEvent(anchor="unknown")
        result = rewriter.resolve_alias(alias)
        assert result == alias

    def test_resolve_alias_invalid_type(self):
        """Test alias resolution when anchor points to non-scalar."""
        yaml_content = "test: value"
        loader = BaseLoader(yaml_content)
        rewriter = Rewriter(loader)
        
        # Mock an anchor that points to something other than ScalarEvent
        from yaml import MappingStartEvent, AliasEvent
        non_scalar_event = MappingStartEvent(anchor=None, tag=None, implicit=True)
        rewriter.anchors["bad_anchor"] = non_scalar_event # type: ignore
        
        alias = AliasEvent(anchor="bad_anchor")
        with pytest.raises(ValueError, match="does not point to a valid scalar event"):
            rewriter.resolve_alias(alias)

class TestTemplate:
    def test_template_initialization(self):
        """Test Template object initialization."""
        template = Template("fake.name()", "test_var")
        assert template.eval_str == "fake.name()"
        assert template.parent_variable == "test_var"

    def test_template_eval_success(self):
        """Test successful template evaluation."""
        template = Template("fake.name()", "test_var")
        result = template.eval()
        assert isinstance(result, str)
        assert len(result) > 0

    def test_template_eval_empty(self):
        """Test template evaluation with empty string."""
        template = Template("", "test_var")
        result = template.eval()
        assert result is None

    def test_template_eval_invalid_syntax(self):
        """Test template evaluation with invalid Python syntax."""
        template = Template("fake.invalid_function(", "test_var")  # missing closing paren
        with pytest.raises(ValueError, match="Invalid template"):
            template.eval()

    def test_template_eval_nonexistent_function(self):
        """Test template evaluation with non-existent function."""
        template = Template("fake.nonexistent_method()", "test_var")
        with pytest.raises(ValueError, match="Invalid template"):
            template.eval()

    def test_template_eval_invalid_code(self):
        """Test template evaluation with malicious/invalid code."""
        template = Template("import os; os.system('echo bad')", "test_var")
        with pytest.raises(ValueError, match="Invalid template"):
            template.eval()

    def test_template_from_yaml_invalid_node_type(self):
        """Test Template.from_yaml with invalid node type."""
        from yaml.nodes import ScalarNode
        node = ScalarNode(tag="tag", value="value")
        with pytest.raises(yaml.YAMLError, match="Template must be a mapping"):
            Template.from_yaml(None, node)

    def test_template_from_yaml_missing_eval_key(self):
        """Test Template.from_yaml with missing 'eval' key."""
        from yaml.nodes import MappingNode, ScalarNode
        
        node = MappingNode(tag="tag", value=[
            (ScalarNode(tag="tag", value="variable"), ScalarNode(tag="tag", value="test"))
        ])
        with pytest.raises(yaml.YAMLError, match="Template must have a 'eval' key"):
            Template.from_yaml(None, node)

    def test_template_from_yaml_missing_variable_key(self):
        """Test Template.from_yaml with missing 'variable' key."""
        from yaml.nodes import MappingNode, ScalarNode
        
        node = MappingNode(tag="tag", value=[
            (ScalarNode(tag="tag", value="eval"), ScalarNode(tag="tag", value="fake.name()"))
        ])
        with pytest.raises(yaml.YAMLError, match="Template must have a 'variable' key"):
            Template.from_yaml(None, node)

    def test_template_from_yaml_missing_both_keys(self):
        """Test Template.from_yaml with both required keys missing."""
        from yaml.nodes import MappingNode, ScalarNode
        
        node = MappingNode(tag="tag", value=[
            (ScalarNode(tag="tag", value="other"), ScalarNode(tag="tag", value="value"))
        ])
        with pytest.raises(yaml.YAMLError, match="Template must have a 'eval' key"):
            Template.from_yaml(None, node)

    def test_template_yaml_roundtrip(self):
        """Test Template YAML serialization and deserialization."""
        template = Template("fake.name()", "test_var")
        
        # Test to_yaml
        dumper = yaml.Dumper(io.StringIO())
        node = Template.to_yaml(dumper, template)
        assert node.tag == "!template"
        
        # Test from_yaml
        reconstructed = Template.from_yaml(None, node)
        assert reconstructed.eval_str == template.eval_str
        assert reconstructed.parent_variable == template.parent_variable

    def test_template_from_yaml_with_extra_keys(self):
        """Test Template.from_yaml ignores extra keys but requires eval and variable."""
        from yaml.nodes import MappingNode, ScalarNode
        
        node = MappingNode(tag="tag", value=[
            (ScalarNode(tag="tag", value="eval"), ScalarNode(tag="tag", value="fake.name()")),
            (ScalarNode(tag="tag", value="variable"), ScalarNode(tag="tag", value="test_var")),
            (ScalarNode(tag="tag", value="extra"), ScalarNode(tag="tag", value="ignored"))
        ])
        
        template = Template.from_yaml(None, node)
        assert template.eval_str == "fake.name()"
        assert template.parent_variable == "test_var"

class TestRewriterIntegration:
    def test_rewrite_valid_variable_with_alias(self):
        """Test complete rewriting workflow with valid variable."""
        yaml_content = """
variables:
  flag:
    template: "fake.bothify('CTF{????-####}')"
    default: &flag_var "default_flag"
services:
  app:
    environment:
      FLAG: *flag_var
"""
        loader = BaseLoader(yaml_content)
        rewriter = Rewriter(loader)
        
        events = list(rewriter.rewrite())
        assert len(events) > 0
        
        # The events should contain the rewritten template
        event_values = [e.value if hasattr(e, 'value') else str(e) for e in events] # type: ignore
        assert 'variables' in event_values


# Copyright 2025 Cyber Skyline

# Permission is hereby granted, free of charge, to any person obtaining a 
# copy of this software and associated documentation files (the “Software”), 
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
# IN THE SOFTWARE.
from cyber_skyline.chall_parser.compose.challenge_info import ChallengeInfo, Hint, Question, TextBody, Variable
from cyber_skyline.chall_parser.template import Template

class TestQuestion:
    def test_basic_question(self):
        """Test creating a basic question."""
        question = Question(
            name="flag",
            body="What is the flag?",
            points=100,
            answer="CTF{test}",
            max_attempts=5
        )
        assert question.name == "flag"
        assert question.points == 100
        assert question.max_attempts == 5

class TestHint:
    def test_text_hint(self):
        """Test creating a text hint."""
        text_hint = TextBody(type="text", content="This is a hint")
        hint = Hint(
            name="hint1",
            body=text_hint,
            preview="Hint preview",
            deduction=10
        )
        assert isinstance(hint.body, TextBody)
        assert hint.body.content == "This is a hint"
        assert hint.deduction == 10

    def test_string_hint(self):
        """Test creating a simple string hint."""
        hint = Hint(
            name="hint2",
            body="Simple hint text",
            preview="Simple hint",
            deduction=5
        )
        assert isinstance(hint.body, str)
        assert hint.body == "Simple hint text"

class TestVariable:
    def test_variable_with_template(self):
        """Test creating a variable with template."""
        template = Template("fake.word()", "test_var")
        variable = Variable(
            template=template,
            default="default_value"
        )
        assert variable.template == template
        assert variable.default == "default_value"

class TestChallengeInfo:
    def test_minimal_challenge_info(self):
        """Test creating minimal challenge info."""
        challenge = ChallengeInfo(
            name="Test Challenge",
            description="A test challenge",
            questions=[]
        )
        assert challenge.name == "Test Challenge"
        assert challenge.description == "A test challenge"
        assert len(challenge.questions) == 0

    def test_challenge_info_with_all_optional_fields(self):
        """Test ChallengeInfo with all optional fields populated."""
        question = Question(
            name="comprehensive_test",
            body="Test question?",
            points=50,
            answer="test_answer",
            max_attempts=10,
            placeholder="Enter your answer here"  # Optional field added
        )
        
        hint = Hint(
            name="hint1",
            body=TextBody(type="text", content="Test hint content"),
            preview="Test preview",
            deduction=5
        )
        
        variable = Variable(
            template=Template("fake.word()", "test_var"),
            default="default_value"
        )
        
        challenge = ChallengeInfo(
            name="Comprehensive Test",
            description="Testing all fields", 
            questions=[question],
            icon="TbTest",
            hints=[hint],
            summary="Test summary",
            templates={"test_template": "fake.name()"},
            variables={"test_var": variable},
            tags=["test", "comprehensive"]
        )
        
        assert challenge.icon == "TbTest"
        assert challenge.summary == "Test summary"
        assert challenge.templates is not None
        assert "test_template" in challenge.templates
        assert challenge.variables is not None
        assert "test_var" in challenge.variables
        assert challenge.tags is not None
        assert "test" in challenge.tags

    def test_challenge_info_edge_cases(self):
        """Test ChallengeInfo with edge cases."""
        # Empty lists
        challenge = ChallengeInfo(
            name="Edge Case Test",
            description="Testing edge cases",
            questions=[],
            hints=[],
            tags=[]
        )
        
        assert len(challenge.questions) == 0
        assert challenge.hints is not None
        assert len(challenge.hints) == 0
        assert challenge.tags is not None
        assert len(challenge.tags) == 0